@@ -234,7 +234,9 @@ int spider_udf_direct_sql_create_conn_key(
       + 1 + 2
       + (direct_sql->tgt_default_file ? direct_sql->tgt_default_file_length + 2 : 0)
       + (direct_sql->tgt_default_group ? direct_sql->tgt_default_group_length + 2 : 0)
-      + (direct_sql->tgt_dsn ? direct_sql->tgt_dsn_length + 2 : 0);
+      + (direct_sql->tgt_dsn ? direct_sql->tgt_dsn_length + 2 : 0)
+      + (direct_sql->tgt_filedsn ? direct_sql->tgt_filedsn_length + 2 : 0)
+      + (direct_sql->tgt_driver ? direct_sql->tgt_driver_length + 2 : 0);
   if (!(direct_sql->conn_key = (char *)
     spider_malloc(spider_current_trx, SPD_MID_UDF_DIRECT_SQL_CREATE_CONN_KEY_1, direct_sql->conn_key_length + 1,
       MYF(MY_WME | MY_ZEROFILL)))
@@ -273,6 +275,8 @@ int spider_udf_direct_sql_create_conn_key(
   spider_create_conn_key_add_one(&counter, &tmp_name, direct_sql->tgt_default_file);
   spider_create_conn_key_add_one(&counter, &tmp_name, direct_sql->tgt_default_group);
   spider_create_conn_key_add_one(&counter, &tmp_name, direct_sql->tgt_dsn);
+  spider_create_conn_key_add_one(&counter, &tmp_name, direct_sql->tgt_filedsn);
+  spider_create_conn_key_add_one(&counter, &tmp_name, direct_sql->tgt_driver);
   tmp_name++;
   direct_sql->conn_key_hash_value = my_calc_hash(&spider_open_connections,
     (uchar*) direct_sql->conn_key, direct_sql->conn_key_length);
@@ -305,7 +309,7 @@ SPIDER_CONN *spider_udf_direct_sql_create_conn(
   char *tmp_name, *tmp_host, *tmp_username, *tmp_password, *tmp_socket;
   char *tmp_wrapper, *tmp_db, *tmp_ssl_ca, *tmp_ssl_capath, *tmp_ssl_cert;
   char *tmp_ssl_cipher, *tmp_ssl_key, *tmp_default_file, *tmp_default_group;
-  char *tmp_dsn;
+  char *tmp_dsn, *tmp_filedsn, *tmp_driver;
   int *need_mon;
   bool tables_on_different_db_are_joinable = TRUE;
   DBUG_ENTER("spider_udf_direct_sql_create_conn");
@@ -342,6 +346,10 @@ SPIDER_CONN *spider_udf_direct_sql_create_conn(
           (uint) (direct_sql->tgt_default_group_length + 1),
         &tmp_dsn,
           (uint) (direct_sql->tgt_dsn_length + 1),
+        &tmp_filedsn,
+          (uint) (direct_sql->tgt_filedsn_length + 1),
+        &tmp_driver,
+          (uint) (direct_sql->tgt_driver_length + 1),
         &need_mon, (uint) (sizeof(int)),
         NullS))
     ) {
@@ -397,6 +405,12 @@ SPIDER_CONN *spider_udf_direct_sql_create_conn(
     spider_maybe_memcpy_string(
       &conn->tgt_dsn, direct_sql->tgt_dsn, tmp_dsn,
       &conn->tgt_dsn_length, direct_sql->tgt_dsn_length);
+    spider_maybe_memcpy_string(
+      &conn->tgt_filedsn, direct_sql->tgt_filedsn, tmp_filedsn,
+      &conn->tgt_filedsn_length, direct_sql->tgt_filedsn_length);
+    spider_maybe_memcpy_string(
+      &conn->tgt_driver, direct_sql->tgt_driver, tmp_driver,
+      &conn->tgt_driver_length, direct_sql->tgt_driver_length);
     conn->tgt_ssl_vsc = direct_sql->tgt_ssl_vsc;
   conn->dbton_id = direct_sql->dbton_id;
   conn->conn_need_mon = need_mon;
@@ -416,6 +430,7 @@ SPIDER_CONN *spider_udf_direct_sql_create_conn(
   conn->semi_trx_isolation = -2;
   conn->semi_trx_isolation_chk = FALSE;
   conn->semi_trx_chk = FALSE;
+    conn->conn_kind = SPIDER_CONN_KIND_MYSQL;
 
   if (mysql_mutex_init(spd_key_mutex_mta_conn, &conn->mta_conn_mutex,
     MY_MUTEX_INIT_FAST))
@@ -560,6 +575,7 @@ SPIDER_CONN *spider_udf_direct_sql_get_conn(
     conn->queued_ping = FALSE;
 
   DBUG_PRINT("info",("spider conn=%p", conn));
+  DBUG_PRINT("info",("spider conn->conn_kind=%u", conn->conn_kind));
   DBUG_RETURN(conn);
 
 error:
@@ -836,7 +852,9 @@ int spider_udf_parse_direct_sql_param(
         SPIDER_PARAM_INT("cto", connect_timeout, 0);
         SPIDER_PARAM_STR("dff", tgt_default_file);
         SPIDER_PARAM_STR("dfg", tgt_default_group);
+        SPIDER_PARAM_STR("drv", tgt_driver);
         SPIDER_PARAM_STR("dsn", tgt_dsn);
+        SPIDER_PARAM_STR("fds", tgt_filedsn);
         SPIDER_PARAM_LONGLONG("prt", priority, 0);
         SPIDER_PARAM_INT("rto", net_read_timeout, 0);
         SPIDER_PARAM_STR("sca", tgt_ssl_ca);
@@ -859,6 +877,7 @@ int spider_udf_parse_direct_sql_param(
         error_num= parse.fail(true);
         goto error;
       case 6:
+        SPIDER_PARAM_STR("driver", tgt_driver);
         SPIDER_PARAM_STR("server", server_name);
         SPIDER_PARAM_STR("socket", tgt_socket);
         SPIDER_PARAM_HINT_WITH_MAX("iop", iop, 3, direct_sql->table_count, 0, 2);
@@ -866,6 +885,7 @@ int spider_udf_parse_direct_sql_param(
         error_num= parse.fail(true);
         goto error;
       case 7:
+        SPIDER_PARAM_STR("filedsn", tgt_filedsn);
         SPIDER_PARAM_STR("wrapper", tgt_wrapper);
         SPIDER_PARAM_STR("ssl_key", tgt_ssl_key);
         error_num= parse.fail(true);
@@ -944,8 +964,14 @@ int spider_udf_set_direct_sql_param_default(
 ) {
   bool check_socket;
   bool check_database;
+  bool check_default_file;
+  bool check_host;
+  bool check_port;
   bool socket_has_default_value;
   bool database_has_default_value;
+  bool default_file_has_default_value;
+  bool host_has_default_value;
+  bool port_has_default_value;
   int error_num, roop_count;
   DBUG_ENTER("spider_udf_set_direct_sql_param_default");
   if (direct_sql->server_name)
@@ -968,10 +994,35 @@ int spider_udf_set_direct_sql_param_default(
   } else {
     check_database = FALSE;
   }
-  if (check_socket || check_database)
+  if (
+    !direct_sql->tgt_default_file &&
+    direct_sql->tgt_default_group &&
+    (*spd_defaults_file || *spd_defaults_extra_file)
+  ) {
+    check_default_file = TRUE;
+  } else {
+    check_default_file = FALSE;
+  }
+  if (!direct_sql->tgt_host)
+  {
+    check_host = TRUE;
+  } else {
+    check_host = FALSE;
+  }
+  if (direct_sql->tgt_port == -1)
+  {
+    check_port = TRUE;
+  } else {
+    check_port = FALSE;
+  }
+  if (check_socket || check_database || check_default_file || check_host ||
+    check_port)
   {
     socket_has_default_value = check_socket;
     database_has_default_value = check_database;
+    default_file_has_default_value = check_default_file;
+    host_has_default_value = check_host;
+    port_has_default_value = check_port;
     if (direct_sql->tgt_wrapper)
     {
       for (roop_count = 0; roop_count < SPIDER_DBTON_SIZE; roop_count++)
@@ -999,6 +1050,21 @@ int spider_udf_set_direct_sql_param_default(
               database_has_default_value = spider_dbton[roop_count].
                 db_util->database_has_default_value();
             }
+            if (check_default_file)
+            {
+              default_file_has_default_value = spider_dbton[roop_count].
+                db_util->default_file_has_default_value();
+            }
+            if (check_host)
+            {
+              host_has_default_value = spider_dbton[roop_count].
+                db_util->host_has_default_value();
+            }
+            if (check_port)
+            {
+              port_has_default_value = spider_dbton[roop_count].
+                db_util->port_has_default_value();
+            }
             break;
           }
         }
@@ -1007,6 +1073,9 @@ int spider_udf_set_direct_sql_param_default(
   } else {
     socket_has_default_value = FALSE;
     database_has_default_value = FALSE;
+    default_file_has_default_value = FALSE;
+    host_has_default_value = FALSE;
+    port_has_default_value = FALSE;
   }
 
   if (database_has_default_value)
@@ -1037,7 +1106,7 @@ int spider_udf_set_direct_sql_param_default(
     }
   }
 
-  if (!direct_sql->tgt_host)
+  if (host_has_default_value)
   {
     DBUG_PRINT("info",("spider create default tgt_host"));
     direct_sql->tgt_host_length = strlen(my_localhost);
@@ -1051,11 +1120,8 @@ int spider_udf_set_direct_sql_param_default(
     }
   }
 
-  if (
-    !direct_sql->tgt_default_file &&
-    direct_sql->tgt_default_group &&
-    (*spd_defaults_file || *spd_defaults_extra_file)
-  ) {
+  if (default_file_has_default_value)
+  {
     DBUG_PRINT("info",("spider create default tgt_default_file"));
     if (*spd_defaults_extra_file)
     {
@@ -1082,7 +1148,7 @@ int spider_udf_set_direct_sql_param_default(
   }
 
 
-  if (direct_sql->tgt_port == -1)
+  if (port_has_default_value)
   {
       direct_sql->tgt_port = MYSQL_PORT;
   }
@@ -1208,6 +1274,14 @@ void spider_udf_free_direct_sql_alloc(
   {
     spider_free(spider_current_trx, direct_sql->tgt_dsn, MYF(0));
   }
+  if (direct_sql->tgt_filedsn)
+  {
+    spider_free(spider_current_trx, direct_sql->tgt_filedsn, MYF(0));
+  }
+  if (direct_sql->tgt_driver)
+  {
+    spider_free(spider_current_trx, direct_sql->tgt_driver, MYF(0));
+  }
   if (direct_sql->conn_key)
   {
     spider_free(spider_current_trx, direct_sql->conn_key, MYF(0));
@@ -1223,8 +1297,8 @@ void spider_udf_free_direct_sql_alloc(
 long long spider_direct_sql_body(
   UDF_INIT *initid,
   UDF_ARGS *args,
-  unsigned char *is_null,
-  unsigned char *error,
+  char *is_null,
+  char *error,
   my_bool bg
 ) {
   int error_num, roop_count;
